# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import abc
import dataclasses
import functools
from collections.abc import Mapping, Sequence
from typing import Any, Callable, Final, Generic, TypeAlias, TypeVar, cast

import array_api_compat

import gt4py.eve as eve
import gt4py.next.common as common
import gt4py.next.custom_layout_allocators as next_allocators
import gt4py.next.embedded.nd_array_field as nd_array_field
from gt4py._core import (
    definitions as core_defs,
    ndarray_utils as core_ndarray_utils,
    types as core_types,
)
from gt4py.eve import extended_typing as xtyping


"""
Field construction API.

The user-facing API consists of the functions 'empty', 'zeros', 'ones', 'full' and 'as_field',
or the 'FieldConstructor' class for more advanced use cases.

These functions create GT4Py 'Field's backed by arrays (NDArrayField)
created using a specified allocator, which can be either an array namespace
(e.g. 'numpy', 'cupy') or a GT4Py field buffer allocator (e.g. a backend).

This module deals with 3 concepts:
- allocating with a backing array API or 'FieldBufferAllocator'
- translating from absolute (domain) to relative (shape) indexing, e.g. for handling of 'aligned_index'
- translating from GT4Py dtypes and devices to the format expected by the backing array API.
"""

# Type to be used by the end-user
Allocator: TypeAlias = core_ndarray_utils.ArrayNamespace | next_allocators.FieldBufferAllocationUtil
"""
Type for field memory allocators.

Accepts either:
- An array namespace following the Array API standard (e.g. ``numpy``, ``cupy``, ``jax.numpy``),
  which will be used directly for array creation.
- A GT4Py field buffer allocator or allocator factory (e.g. a backend), which controls
  memory layout and alignment.
"""

DEFAULT_DEVICE: Final[core_defs.Device] = core_defs.Device(core_defs.DeviceType.CPU, 0)
DEFAULT_DTYPE: Final[core_defs.DType] = core_defs.Float64DType(())


class FieldConstructor:
    """
    Public-facing field construction API.

    Delegates array creation to the composed `_array_constructor` and wraps the result in a Field.
    """

    _array_constructor: _FieldArrayConstructor

    def __init__(
        self,
        allocator: Allocator | None = None,
        aligned_index: Sequence[common.NamedIndex] | None = None,
        device: core_defs.Device | None = None,
    ):
        if allocator is None:
            if device is None:
                device = DEFAULT_DEVICE
            # Currently we use optimized CPU layout, but possibly we could also use standard numpy allocation
            allocator = next_allocators.device_allocators[device.device_type]

        if core_ndarray_utils.is_array_namespace(allocator):
            if aligned_index is not None:
                import warnings

                warnings.warn(
                    "`aligned_index` is not supported when using an array namespace allocator and will be ignored.",
                    stacklevel=2,
                )
            translated_device = (
                core_ndarray_utils.get_device_translator(allocator)(device)
                if device is not None
                else None
            )
            self._array_constructor = _ArrayAPIArrayConstructor(
                array_ns=allocator, device=translated_device
            )
        elif next_allocators.is_field_allocation_tool(allocator):
            if next_allocators.is_field_allocator_factory(allocator):
                allocator = allocator.__gt_allocator__
            if device is not None and allocator.__gt_device_type__ != device.device_type:
                raise ValueError(
                    f"Allocator {allocator} is for device type {allocator.__gt_device_type__}, but device type {device.device_type} was specified."
                )
            self._array_constructor = _CustomLayoutConstructor(
                allocator=allocator,
                device=device,
                aligned_index=aligned_index,
            )
        else:
            raise ValueError(f"Invalid field allocator: {allocator}.")

    def _construct_from_array(
        self,
        domain: common.DomainLike,
        dtype: core_defs.DTypeLike,
        ndarray_constructor: Callable,
    ) -> nd_array_field.NdArrayField:
        domain = common.domain(domain)
        dtype = core_defs.dtype(dtype)
        res = common._field(ndarray_constructor(domain, dtype=dtype), domain=domain)
        assert isinstance(res, nd_array_field.NdArrayField)
        return res

    def empty(
        self,
        domain: common.DomainLike,
        *,
        dtype: core_defs.DTypeLike = DEFAULT_DTYPE,
    ) -> nd_array_field.NdArrayField:
        """Create a `Field` of uninitialized values. See :func:`empty` for details."""
        return self._construct_from_array(domain, dtype, self._array_constructor.empty)

    def zeros(
        self,
        domain: common.DomainLike,
        *,
        dtype: core_defs.DTypeLike = DEFAULT_DTYPE,
    ) -> nd_array_field.NdArrayField:
        """Create a `Field` containing all zeros. See :func:`zeros` for details."""
        return self._construct_from_array(domain, dtype, self._array_constructor.zeros)

    def ones(
        self,
        domain: common.DomainLike,
        *,
        dtype: core_defs.DTypeLike = DEFAULT_DTYPE,
    ) -> nd_array_field.NdArrayField:
        """Create a `Field` containing all ones. See :func:`ones` for details."""
        return self._construct_from_array(domain, dtype, self._array_constructor.ones)

    def full(
        self,
        domain: common.DomainLike,
        fill_value: core_defs.Scalar,
        *,
        dtype: core_defs.DTypeLike | None = None,
    ) -> nd_array_field.NdArrayField:
        """Create a `Field` filled with `fill_value`. See :func:`full` for details."""
        dtype = dtype if dtype is not None else type(fill_value)
        return self._construct_from_array(
            domain,
            dtype,
            lambda domain, dtype: self._array_constructor.full(
                domain, fill_value=fill_value, dtype=dtype
            ),
        )

    def as_field(
        self,
        domain: common.DomainLike | Sequence[common.Dimension],
        data: core_defs.NDArrayObject,
        *,
        dtype: core_defs.DTypeLike | None = None,
        origin: Mapping[common.Dimension, int] | None = None,
    ) -> nd_array_field.NdArrayField:
        """Create a `Field` from an array-like object. See :func:`as_field` for details."""
        if isinstance(domain, Sequence) and all(
            isinstance(dim, common.Dimension) for dim in domain
        ):
            domain = cast(Sequence[common.Dimension], domain)
            if len(domain) != data.ndim:
                raise ValueError(
                    f"Cannot construct 'Field' from array of shape '{data.shape}' and domain '{domain}'."
                )
            if origin:
                domain_dims = set(domain)
                if unknown_dims := set(origin.keys()) - domain_dims:
                    raise ValueError(f"Origin keys {unknown_dims} not in domain {domain}.")
            else:
                origin = {}
            actual_domain = common.domain(
                [
                    (d, (-(start_offset := origin.get(d, 0)), s - start_offset))
                    for d, s in zip(domain, data.shape)
                ]
            )
        else:
            if origin:
                raise ValueError(f"Cannot specify origin for a concrete domain {domain}")
            actual_domain = common.domain(cast(common.DomainLike, domain))

        if data.shape != actual_domain.shape:
            raise ValueError(f"Cannot construct 'Field' from array of shape '{data.shape}'.")

        if dtype is None:
            dtype = core_defs.dtype(data.dtype)

        return self._construct_from_array(
            actual_domain,
            dtype,
            lambda domain, dtype: self._array_constructor.asarray(domain, data, dtype=dtype),
        )


class _FieldArrayConstructor(abc.ABC):
    """
    Abstract interface for array creation operations on a specific device/backend.

    The translation from DomainLike and DTypeLike is already done in the public interface.
    """

    @abc.abstractmethod
    def empty(
        self, domain: common.Domain, *, dtype: core_defs.DType
    ) -> core_defs.NDArrayObject: ...

    @abc.abstractmethod
    def zeros(
        self, domain: common.Domain, *, dtype: core_defs.DType
    ) -> core_defs.NDArrayObject: ...

    @abc.abstractmethod
    def ones(self, domain: common.Domain, *, dtype: core_defs.DType) -> core_defs.NDArrayObject: ...

    @abc.abstractmethod
    def full(
        self,
        domain: common.Domain,
        fill_value: core_defs.Scalar,
        *,
        dtype: core_defs.DType,
    ) -> core_defs.NDArrayObject: ...

    @abc.abstractmethod
    def asarray(
        self,
        domain: common.Domain,
        data: core_defs.NDArrayObject,
        *,
        dtype: core_defs.DType,
    ) -> core_defs.NDArrayObject: ...


_ArrayNST = TypeVar("_ArrayNST", bound=core_ndarray_utils.ArrayNamespace)


@dataclasses.dataclass(frozen=True)
class _ArrayAPIArrayConstructor(_FieldArrayConstructor, Generic[_ArrayNST]):
    array_ns: _ArrayNST
    # device in the format expected by the array namespace
    device: Any = None

    def __post_init__(self) -> None:
        object.__setattr__(
            self, "array_ns", array_api_compat.array_namespace(self.array_ns.empty((1,)))
        )

    def _to_array_ns_dtype(self, dtype: core_defs.DType) -> Any:
        return getattr(self.array_ns, core_types.type_to_name[dtype.scalar_type])

    def empty(
        self,
        domain: common.Domain,
        *,
        dtype: core_defs.DType,
    ) -> core_defs.NDArrayObject:
        return self.array_ns.empty(
            domain.shape, dtype=self._to_array_ns_dtype(dtype), device=self.device
        )

    def zeros(
        self,
        domain: common.Domain,
        *,
        dtype: core_defs.DType,
    ) -> core_defs.NDArrayObject:
        return self.array_ns.zeros(
            domain.shape, dtype=self._to_array_ns_dtype(dtype), device=self.device
        )

    def ones(
        self,
        domain: common.Domain,
        *,
        dtype: core_defs.DType,
    ) -> core_defs.NDArrayObject:
        return self.array_ns.ones(
            domain.shape, dtype=self._to_array_ns_dtype(dtype), device=self.device
        )

    def full(
        self,
        domain: common.Domain,
        fill_value: core_defs.Scalar,
        *,
        dtype: core_defs.DType,
    ) -> core_defs.NDArrayObject:
        return self.array_ns.full(
            domain.shape,
            fill_value=fill_value,
            dtype=self._to_array_ns_dtype(dtype),
            device=self.device,
        )

    def asarray(
        self,
        domain: common.Domain,
        data: core_defs.NDArrayObject,
        *,
        dtype: core_defs.DType,
    ) -> core_defs.NDArrayObject:
        arr = self.array_ns.asarray(
            data, dtype=self._to_array_ns_dtype(dtype), device=self.device, copy=True
        )
        assert domain.shape == arr.shape, (
            f"pre-condition of `asarray` not met: {data.shape=} and {domain.shape} need to agree."
        )
        return arr


@dataclasses.dataclass(frozen=True)
class _CustomLayoutConstructor(_FieldArrayConstructor):
    allocator: next_allocators.FieldBufferAllocatorProtocol
    device: core_defs.Device | None = None
    aligned_index: Sequence[common.NamedIndex] | None = None

    @functools.cached_property
    def device_id(self) -> int:
        return self.device.device_id if self.device is not None else 0

    def empty(self, domain: common.Domain, *, dtype: core_defs.DType) -> core_defs.NDArrayObject:
        return self.allocator.__gt_allocate__(
            domain=domain,
            dtype=dtype,
            aligned_index=self.aligned_index,
            device_id=self.device_id,
        ).ndarray

    def zeros(self, domain: common.Domain, *, dtype: core_defs.DType) -> core_defs.NDArrayObject:
        return self.full(domain, fill_value=0, dtype=dtype)

    def ones(self, domain: common.Domain, *, dtype: core_defs.DType) -> core_defs.NDArrayObject:
        return self.full(domain, fill_value=1, dtype=dtype)

    def full(
        self,
        domain: common.Domain,
        fill_value: core_defs.Scalar,
        *,
        dtype: core_defs.DType,
    ) -> core_defs.NDArrayObject:
        arr = self.empty(domain, dtype=dtype)
        arr[...] = fill_value  # type: ignore[index] # `NDArrayObject` typing is not complete
        return arr

    def asarray(
        self,
        domain: common.Domain,
        data: core_defs.NDArrayObject,
        *,
        dtype: core_defs.DType,
    ) -> core_defs.NDArrayObject:
        arr = self.empty(domain, dtype=dtype)

        arr[...] = core_ndarray_utils.array_namespace(arr).asarray(data)  # type: ignore[index] # `NDArrayObject` typing is not complete
        return arr


# Our typical `Allocator`s are hashable, but in case they are not we fall-back to non cached.
# E.g. it's up to the user to make types implementing `__gt_allocator__` hashable.
@eve.utils.optional_lru_cache
def _field_constructor(
    allocator: Allocator | None,
    aligned_index: Sequence[common.NamedIndex] | None = None,
    device: core_defs.Device | None = None,
) -> FieldConstructor:
    return FieldConstructor(allocator, aligned_index=aligned_index, device=device)


@eve.utils.with_fluid_partial
def empty(
    domain: common.DomainLike,
    dtype: core_defs.DTypeLike = DEFAULT_DTYPE,
    *,
    aligned_index: Sequence[common.NamedIndex] | None = None,
    allocator: Allocator | None = None,
    device: core_defs.Device | None = None,
) -> nd_array_field.NdArrayField:
    """Create a `Field` of uninitialized (undefined) values using the given (or device-default) allocator.

    This function supports partial binding of arguments, see :class:`eve.utils.partial` for details.

    Arguments:
        domain: Definition of the domain of the field (which fix the shape of the allocated field buffer).
            See :class:`gt4py.next.common.Domain` for details.
        dtype: Definition of the data type of the field. Defaults to `float64`.

    Keyword Arguments:
        aligned_index: Index in the definition domain which should be used as reference
            point for memory alignment computations. It can be set to the most common origin
            of computations in this domain (if known) for performance reasons.
        allocator: An array namespace (e.g. ``numpy``, ``cupy``) or an allocator /
            allocator factory (e.g. backend) used for memory buffer allocation.
            If neither `allocator` nor `device` is given, a default CPU allocator is used.
            If both are given, `allocator` takes precedence over the default device allocator.
        device: The device (CPU, type of accelerator) to optimize the memory layout for.
            If neither `allocator` nor `device` is given, a default CPU device is used.
            When only `device` is given, the default allocator for that device type is used.

    Returns:
        A field, backed by a buffer with memory layout as specified by allocator and alignment requirements.

    Raises:
        ValueError
            If illegal or inconsistent arguments are specified.

    Examples:
        Initialize a field in one dimension with a backend and a range domain:

        >>> from gt4py import next as gtx
        >>> IDim = gtx.Dimension("I")
        >>> a = gtx.empty({IDim: range(3, 10)}, allocator=gtx.itir_python)
        >>> a.shape
        (7,)

        Initialize a field in one dimension from an array namespace:

        >>> import numpy as np
        >>> from gt4py import next as gtx
        >>> IDim = gtx.Dimension("I")
        >>> a = gtx.empty({IDim: range(3, 10)}, allocator=np)
        >>> a.shape
        (7,)

        Initialize with a device and an integer domain. It works like a shape with named dimensions:

        >>> from gt4py._core import definitions as core_defs
        >>> JDim = gtx.Dimension("J")
        >>> b = gtx.empty(
        ...     {IDim: 3, JDim: 3}, int, device=core_defs.Device(core_defs.DeviceType.CPU, 0)
        ... )
        >>> b.shape
        (3, 3)
    """
    return _field_constructor(allocator, aligned_index=aligned_index, device=device).empty(
        domain, dtype=dtype
    )


@eve.utils.with_fluid_partial
def zeros(
    domain: common.DomainLike,
    dtype: core_defs.DTypeLike = DEFAULT_DTYPE,
    *,
    aligned_index: Sequence[common.NamedIndex] | None = None,
    allocator: Allocator | None = None,
    device: core_defs.Device | None = None,
) -> nd_array_field.NdArrayField:
    """Create a Field containing all zeros using the given (or device-default) allocator.

    This function supports partial binding of arguments, see :class:`eve.utils.partial` for details.
    See :func:`empty` for further details about the meaning of the arguments.

    Examples:
        >>> from gt4py import next as gtx
        >>> IDim = gtx.Dimension("I")
        >>> gtx.zeros({IDim: range(3, 10)}, allocator=gtx.itir_python).ndarray
        array([0., 0., 0., 0., 0., 0., 0.])
    """
    return _field_constructor(allocator, aligned_index=aligned_index, device=device).zeros(
        domain, dtype=dtype
    )


@eve.utils.with_fluid_partial
def ones(
    domain: common.DomainLike,
    dtype: core_defs.DTypeLike = DEFAULT_DTYPE,
    *,
    aligned_index: Sequence[common.NamedIndex] | None = None,
    allocator: Allocator | None = None,
    device: core_defs.Device | None = None,
) -> nd_array_field.NdArrayField:
    """Create a Field containing all ones using the given (or device-default) allocator.

    This function supports partial binding of arguments, see :class:`eve.utils.partial` for details.
    See :func:`empty` for further details about the meaning of the arguments.

    Examples:
        >>> from gt4py import next as gtx
        >>> IDim = gtx.Dimension("I")
        >>> gtx.ones({IDim: range(3, 10)}, allocator=gtx.itir_python).ndarray
        array([1., 1., 1., 1., 1., 1., 1.])
    """
    return _field_constructor(allocator, aligned_index=aligned_index, device=device).ones(
        domain, dtype=dtype
    )


@eve.utils.with_fluid_partial
def full(
    domain: common.DomainLike,
    fill_value: core_defs.Scalar,
    dtype: core_defs.DTypeLike | None = None,
    *,
    aligned_index: Sequence[common.NamedIndex] | None = None,
    allocator: Allocator | None = None,
    device: core_defs.Device | None = None,
) -> nd_array_field.NdArrayField:
    """Create a Field where all values are set to `fill_value` using the given (or device-default) allocator.

    This function supports partial binding of arguments, see :class:`eve.utils.partial` for details.
    See :func:`empty` for further details about the meaning of the arguments.

    Arguments:
        domain: Definition of the domain of the field (and consequently of the shape of the allocated field buffer).
        fill_value: Each point in the field will be initialized to this value.
        dtype: Definition of the data type of the field. Defaults to the dtype of `fill_value`.

    Examples:
        >>> from gt4py import next as gtx
        >>> IDim = gtx.Dimension("I")
        >>> gtx.full({IDim: 3}, 5, allocator=gtx.itir_python).ndarray
        array([5, 5, 5])
    """
    return _field_constructor(allocator, aligned_index=aligned_index, device=device).full(
        domain, fill_value=fill_value, dtype=dtype
    )


@eve.utils.with_fluid_partial
def as_field(
    domain: common.DomainLike | Sequence[common.Dimension],
    data: core_defs.NDArrayObject,
    dtype: core_defs.DTypeLike | None = None,
    *,
    origin: Mapping[common.Dimension, int] | None = None,
    aligned_index: Sequence[common.NamedIndex] | None = None,
    allocator: Allocator | None = None,
    device: core_defs.Device | None = None,
) -> nd_array_field.NdArrayField:
    """Create a Field from an array-like object using the given (or device-default) allocator.

    This function supports partial binding of arguments, see :class:`eve.utils.partial` for details.
    See :func:`empty` for further details about the meaning of the extra keyword arguments.

    Arguments:
        domain: Definition of the domain of the field (and consequently of the shape of the allocated field buffer).
            In addition to the values allowed in `empty`, it can also just be a sequence of dimensions,
            in which case the sizes of each dimension will then be taken from the shape of `data`.
        data: Array like data object to initialize the field with
        dtype: Definition of the data type of the field. Defaults to the same as `data`.

    Keyword Arguments:
        origin: Only allowed if `domain` is a sequence of dimensions. The indicated index in `data`
            will be the zero point of the resulting field.
        allocator: An array namespace (e.g. ``numpy``) or an allocator / allocator factory (e.g. backend).
            If not given, defaults to the device of `data` (if available) or the default CPU allocator.
        device: The device to optimize memory layout for. If not given, defaults to the device
            of `data` (if available) or the default CPU device.

    Note: we do not support a 'copy' argument as we want to avoid creating fields aliasing other data/fields.

    Examples:
        >>> import numpy as np
        >>> from gt4py import next as gtx
        >>> IDim = gtx.Dimension("I")
        >>> xdata = np.array([1, 2, 3])

        Automatic domain from just dimensions:

        >>> a = gtx.as_field([IDim], xdata)
        >>> a.ndarray
        array([1, 2, 3])
        >>> a.domain.ranges[0]
        UnitRange(0, 3)

        Shifted domain using origin:

        >>> b = gtx.as_field([IDim], xdata, origin={IDim: 1})
        >>> b.domain.ranges[0]
        UnitRange(-1, 2)

        Equivalent domain fully specified:

        >>> gtx.as_field({IDim: range(-1, 2)}, xdata).domain.ranges[0]
        UnitRange(-1, 2)
    """
    if allocator is None and device is None and xtyping.supports_dlpack(data):
        # allocate for the device of the input data if no explicit allocator or device is given
        device = core_defs.Device(*data.__dlpack_device__())
    return _field_constructor(allocator, aligned_index=aligned_index, device=device).as_field(
        domain=domain, data=data, dtype=dtype, origin=origin
    )


@eve.utils.with_fluid_partial
def as_connectivity(
    domain: common.DomainLike | Sequence[common.Dimension],
    codomain: common.Dimension,
    data: core_defs.NDArrayObject,
    dtype: core_defs.DTypeLike | None = None,
    *,
    origin: Mapping[common.Dimension, int] | None = None,
    aligned_index: Sequence[common.NamedIndex] | None = None,
    allocator: Allocator | None = None,
    device: core_defs.Device | None = None,
    skip_value: core_defs.IntegralScalar | eve.NothingType | None = eve.NOTHING,
) -> common.Connectivity:
    """
    Construct a `Connectivity` from the given domain, codomain, and data.

    Arguments:
        domain: The domain of the connectivity. It can be either a `common.DomainLike` object or a
            sequence of `common.Dimension` objects.
        codomain: The codomain dimension of the connectivity.
        data: The data used to construct the connectivity field.
        dtype: The data type of the connectivity. If not provided, it will be inferred from the data.

    Keyword Arguments:
        origin: Only allowed if `domain` is a sequence of dimensions. The indicated index in `data`
            will be the zero point of the resulting connectivity.
        aligned_index: Index in the definition domain which should be used as reference
            point for memory alignment computations.
        allocator: An array namespace or an allocator / allocator factory (e.g. backend).
            If not provided, defaults to the device of `data` or the default CPU allocator.
        device: The device on which the connectivity will be allocated. If not provided,
            defaults to the device of `data` or the default CPU device.
        skip_value: The value that signals missing entries in the neighbor table. Defaults to the default
            skip value if it is found in data, otherwise to `None` (= no skip value).

    Returns:
        The constructed connectivity field.

    Raises:
        ValueError: If the domain or codomain is invalid, or if the shape of the data does not match the domain shape.

    Examples:
        >>> import numpy as np
        >>> from gt4py import next as gtx
        >>> Vertex = gtx.Dimension("Vertex")
        >>> Edge = gtx.Dimension("Edge")
        >>> V2EDim = gtx.Dimension("V2E", kind=gtx.DimensionKind.LOCAL)
        >>> data = np.array([[0, 1], [1, 2], [2, 0]])
        >>> conn = gtx.as_connectivity([Vertex, V2EDim], Edge, data)
        >>> conn.ndarray
        array([[0, 1],
               [1, 2],
               [2, 0]])
        >>> conn.domain
        Domain(dims=(Dimension(value='Vertex', kind=<DimensionKind.HORIZONTAL: 'horizontal'>), Dimension(value='V2E', kind=<DimensionKind.LOCAL: 'local'>)), ranges=(UnitRange(0, 3), UnitRange(0, 2)))
    """
    if skip_value is eve.NOTHING:
        skip_value = (
            common._DEFAULT_SKIP_VALUE if (data == common._DEFAULT_SKIP_VALUE).any() else None
        )

    field = as_field(
        domain=domain,
        data=data,
        dtype=dtype,
        origin=origin,
        aligned_index=aligned_index,
        allocator=allocator,
        device=device,
    )

    assert skip_value is not eve.NOTHING  # for mypy
    connectivity_field = common._connectivity(
        field.ndarray,
        codomain=codomain,
        domain=field.domain,
        skip_value=cast(core_defs.IntegralScalar | None, skip_value),  # see assert above
    )
    assert isinstance(connectivity_field, nd_array_field.NdArrayConnectivityField)

    return connectivity_field



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from typing import Any, Callable, Generic, Optional, ParamSpec, Sequence, TypeVar

from gt4py import eve
from gt4py._core import definitions as core_defs
from gt4py.eve import extended_typing as xtyping
from gt4py.next import common, errors, field_utils, named_collections, utils
from gt4py.next.embedded import common as embedded_common, context as embedded_context
from gt4py.next.field_utils import get_array_ns
from gt4py.next.otf import arguments
from gt4py.next.type_system import type_specifications as ts, type_translation


_P = ParamSpec("_P")
_R = TypeVar("_R")


@dataclasses.dataclass(frozen=True)
class EmbeddedOperator(Generic[_R, _P]):
    fun: Callable[_P, _R]

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        return self.fun(*args, **kwargs)


@dataclasses.dataclass(frozen=True)
class ScanOperator(EmbeddedOperator[xtyping.MaybeNestedInTuple[core_defs.ScalarT], _P]):
    forward: bool
    init: xtyping.MaybeNestedInTuple[core_defs.ScalarT]
    axis: common.Dimension

    def __call__(  # type: ignore[override]
        self,
        *args: common.Field | core_defs.Scalar,
        **kwargs: common.Field | core_defs.Scalar,
    ) -> (
        common.Field[Any, core_defs.ScalarT]
        | tuple[common.Field[Any, core_defs.ScalarT] | tuple, ...]
    ):
        scan_range = embedded_context.get_closure_column_range()
        assert self.axis == scan_range.dim
        scan_axis = scan_range.dim
        all_args = [*args, *kwargs.values()]
        domain_intersection = _intersect_scan_args(*all_args)
        non_scan_domain = common.Domain(*[nr for nr in domain_intersection if nr.dim != scan_axis])

        out_domain = common.Domain(
            *[scan_range if nr.dim == scan_axis else nr for nr in domain_intersection]
        )
        if scan_axis not in out_domain.dims:
            # even if the scan dimension is not in the input, we can scan over it
            out_domain = common.Domain(*out_domain, (scan_range))

        xp = get_array_ns(*(arguments.extract(arg) for arg in all_args))
        init_type = type_translation.from_value(self.init)
        assert isinstance(init_type, ts.TupleType | ts.ScalarType | ts.NamedCollectionType)
        res = field_utils.field_from_typespec(init_type, out_domain, xp)

        def scan_loop(hpos: Sequence[common.NamedIndex]) -> None:
            acc: xtyping.MaybeNestedInTuple[core_defs.ScalarT] = self.init
            for k in scan_range.unit_range if self.forward else reversed(scan_range.unit_range):
                pos = (*hpos, common.NamedIndex(scan_axis, k))
                new_args = [_tuple_at(pos, arg) for arg in args]
                new_kwargs = {k: _tuple_at(pos, v) for k, v in kwargs.items()}
                acc = self.fun(acc, *new_args, **new_kwargs)  # type: ignore[arg-type] # need to express that the first argument is the same type as the return
                # convert custom NamedCollections to plain tuples for assignment
                acc_extracted = arguments.extract(acc)
                res_extracted = arguments.extract(res)
                assert xtyping.is_maybe_nested_in_tuple_of(acc_extracted, core_defs.Scalar)  # type: ignore[arg-type]  # Scalar is a Union
                assert xtyping.is_maybe_nested_in_tuple_of(res_extracted, common.MutableField)  # type: ignore[type-abstract]  # MutableField is abstract/generic
                _tuple_assign_value(pos, res_extracted, acc_extracted)

        if len(non_scan_domain) == 0:
            # if we don't have any dimension orthogonal to scan_axis, we need to do one scan_loop
            scan_loop(())
        else:
            for hpos in embedded_common.iterate_domain(non_scan_domain):
                scan_loop(hpos)

        return res


def _get_out_domain(out: xtyping.MaybeNestedInTuple[common.MutableField]) -> common.Domain:
    return embedded_common.domain_intersection(
        *[f.domain for f in utils.flatten_nested_tuple((out,))]
    )


def field_operator_call(op: EmbeddedOperator[_R, _P], args: Any, kwargs: Any) -> Optional[_R]:
    if "out" in kwargs:
        # called from program or direct field_operator as program
        new_context_kwargs = {}
        if embedded_context.within_valid_context():
            # called from program
            assert "offset_provider" not in kwargs
        else:
            # field_operator as program
            if "offset_provider" not in kwargs:
                raise errors.MissingArgumentError(None, "offset_provider", True)
            offset_provider = kwargs.pop("offset_provider", None)

            new_context_kwargs["offset_provider"] = offset_provider

        out = kwargs.pop("out")

        domain = kwargs.pop("domain", None)

        # TODO(havogt): To do the assignment of the resulting fields we extract containers and act on plain tuples.
        # We currently apply the extract on both the rhs (`res`) computed by the operator and the lhs (`out`, provided by the user)
        # without checking if the types are consistent. However, these errors are caught in linting if enabled.
        container_extracted_out = arguments.extract(out)
        assert xtyping.is_maybe_nested_in_tuple_of(container_extracted_out, common.MutableField)  # type: ignore[type-abstract]  # MutableField is abstract/generic
        out_domain = (
            utils.tree_map(common.domain)(domain)
            if domain is not None
            else _get_out_domain(container_extracted_out)
        )

        new_context_kwargs["closure_column_range"] = _get_vertical_range(out_domain)

        with embedded_context.update(**new_context_kwargs):
            res = op(*args, **kwargs)
        container_extracted_res = arguments.extract(res)  # type: ignore[arg-type] # TODO(havogt): see notes above
        _tuple_assign_field(container_extracted_out, container_extracted_res, domain=out_domain)  # type: ignore[arg-type]
        return None
    else:
        # called from other field_operator or missing `out` argument
        if "offset_provider" in kwargs:
            # assuming we wanted to call the field_operator as program, otherwise `offset_provider` would not be there
            raise errors.MissingArgumentError(None, "out", True)
        return op(*args, **kwargs)


@utils.tree_map
def _get_vertical_range(domain: common.Domain) -> common.NamedRange | eve.NothingType:
    vertical_dim_filtered = [nr for nr in domain if nr.dim.kind == common.DimensionKind.VERTICAL]
    assert len(vertical_dim_filtered) <= 1
    return vertical_dim_filtered[0] if vertical_dim_filtered else eve.NOTHING


def _tuple_assign_field(
    target: xtyping.MaybeNestedInTuple[common.MutableField],
    source: xtyping.MaybeNestedInTuple[common.Field],
    domain: xtyping.MaybeNestedInTuple[common.Domain],
) -> None:
    @named_collections.tree_map_named_collection
    def impl(target: common.MutableField, source: common.Field, domain: common.Domain) -> None:
        if isinstance(source, common.Field):
            target[domain] = source[domain]
        else:
            assert core_defs.is_scalar_type(source)
            target[domain] = source

    if not isinstance(domain, tuple):
        domain = named_collections.tree_map_named_collection(lambda _: domain)(target)  # type: ignore[assignment] # typing not precise enough
    impl(target, source, domain)


def _intersect_scan_args(
    *args: xtyping.MaybeNestedInTuple[core_defs.Scalar | common.Field],
) -> common.Domain:
    return embedded_common.domain_intersection(
        *[arg.domain for arg in utils.flatten_nested_tuple(args) if isinstance(arg, common.Field)]
    )


def _tuple_assign_value(
    pos: Sequence[common.NamedIndex],
    target: xtyping.MaybeNestedInTuple[common.MutableField],
    source: xtyping.MaybeNestedInTuple[core_defs.Scalar],
) -> None:
    @utils.tree_map
    def impl(target: common.MutableField, source: core_defs.Scalar) -> None:
        target[pos] = source

    impl(target, source)


def _tuple_at(
    pos: Sequence[common.NamedIndex],
    field: xtyping.MaybeNestedInTuple[common.Field | core_defs.Scalar],
) -> core_defs.Scalar | tuple[core_defs.ScalarT | tuple, ...]:
    @named_collections.tree_map_named_collection
    def impl(field: common.Field | core_defs.Scalar) -> core_defs.Scalar:
        res = field[pos].as_scalar() if isinstance(field, common.Field) else field
        assert core_defs.is_scalar_type(res)
        return res

    return impl(field)  # type: ignore[return-value]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import os
import warnings
from collections.abc import Callable, MutableSequence, Sequence
from typing import Any

import dace
import factory

from gt4py._core import definitions as core_defs, locking
from gt4py.next import common, config
from gt4py.next.otf import code_specs, definitions, stages, workflow
from gt4py.next.otf.compilation import cache as gtx_cache
from gt4py.next.program_processors.runners.dace.workflow import common as gtx_wfdcommon


class CompiledDaceProgram:
    sdfg_program: dace.CompiledSDFG

    # Sorted list of SDFG arguments as they appear in program ABI and corresponding data type;
    # scalar arguments that are not used in the SDFG will not be present.
    sdfg_argtypes: list[dace.dtypes.Data]

    # The compiled program contains a callable object to update the SDFG arguments list.
    update_sdfg_ctype_arglist: Callable[
        [
            core_defs.DeviceType,
            Sequence[dace.dtypes.Data],
            Sequence[Any],
            MutableSequence[Any],
            common.OffsetProvider,
        ],
        None,
    ]

    # Processed argument vectors that are passed to `CompiledSDFG.fast_call()`. `None`
    #  means that it has not been initialized, i.e. no call was ever performed.
    #  - csdfg_argv: Arguments used for calling the actual compiled SDFG, will be updated.
    #  - csdfg_init_argv: Arguments used for initialization; used only the first time and
    #       never updated.
    csdfg_argv: MutableSequence[Any] | None
    csdfg_init_argv: Sequence[Any] | None

    def __init__(
        self,
        program: dace.CompiledSDFG,
        bind_func_name: str,
        binding_source: stages.BindingSource[code_specs.SDFGCodeSpec, code_specs.PythonCodeSpec],
    ):
        self.sdfg_program = program

        # `dace.CompiledSDFG.arglist()` returns an ordered dictionary that maps the argument
        # name to its data type, in the same order as arguments appear in the program ABI.
        # This is also the same order of arguments in `dace.CompiledSDFG._lastargs[0]`.
        self.sdfg_argtypes = list(program.sdfg.arglist().values())

        # Note that `binding_source` contains Python code tailored to this specific SDFG.
        # Here we dinamically compile this function and add it to the compiled program.
        exec(binding_source.source_code, global_namespace := {})  # type: ignore[var-annotated]
        self.update_sdfg_ctype_arglist = global_namespace[bind_func_name]
        # For debug purpose, we set a unique module name on the compiled function.
        self.update_sdfg_ctype_arglist.__module__ = os.path.basename(program.sdfg.build_folder)

        # Since the SDFG hasn't been called yet.
        self.csdfg_argv = None
        self.csdfg_init_argv = None

    def construct_arguments(self, **kwargs: Any) -> None:
        """
        This function will process the arguments and store the processed argument
        vectors in `self.csdfg_args`, to call them use `self.fast_call()`.
        """
        with dace.config.set_temporary("compiler", "allow_view_arguments", value=True):
            csdfg_argv, csdfg_init_argv = self.sdfg_program.construct_arguments(**kwargs)
        # Note we only care about `csdfg_argv` (normal call), since we have to update it,
        #  we ensure that it is a `list`.
        self.csdfg_argv = [*csdfg_argv]
        self.csdfg_init_argv = csdfg_init_argv

    def fast_call(self) -> None:
        """
        Perform a call to the compiled SDFG using the previously generated argument
        vectors, see `self.construct_arguments()`.
        """
        assert self.csdfg_argv is not None and self.csdfg_init_argv is not None, (
            "Argument vector was not set properly."
        )
        self.sdfg_program.fast_call(
            self.csdfg_argv, self.csdfg_init_argv, do_gpu_check=config.DEBUG
        )

    def __call__(self, **kwargs: Any) -> None:
        """Call the compiled SDFG with the given arguments.

        Note that this function will not update the argument vectors stored inside
        `self`. Furthermore, it is not recommended to use this function as it is
        very slow.
        """
        warnings.warn(
            "Called an SDFG through the standard DaCe interface is not recommended, use `fast_call()` instead.",
            stacklevel=1,
        )
        result = self.sdfg_program(**kwargs)
        assert result is None


@dataclasses.dataclass(frozen=True)
class DaCeCompiler(
    workflow.ChainableWorkflowMixin[
        stages.CompilableProject[code_specs.SDFGCodeSpec, code_specs.PythonCodeSpec],
        CompiledDaceProgram,
    ],
    workflow.ReplaceEnabledWorkflowMixin[
        stages.CompilableProject[code_specs.SDFGCodeSpec, code_specs.PythonCodeSpec],
        CompiledDaceProgram,
    ],
    definitions.CompilationStep[code_specs.SDFGCodeSpec, code_specs.PythonCodeSpec],
):
    """Use the dace build system to compile a GT4Py program to a ``gt4py.next.otf.stages.CompiledProgram``."""

    bind_func_name: str
    cache_lifetime: config.BuildCacheLifetime
    device_type: core_defs.DeviceType
    cmake_build_type: config.CMakeBuildType = config.CMakeBuildType.DEBUG

    def __call__(
        self,
        inp: stages.CompilableProject[code_specs.SDFGCodeSpec, code_specs.PythonCodeSpec],
    ) -> CompiledDaceProgram:
        with gtx_wfdcommon.dace_context(
            device_type=self.device_type,
            cmake_build_type=self.cmake_build_type,
        ):
            sdfg_build_folder = gtx_cache.get_cache_folder(inp, self.cache_lifetime)
            sdfg_build_folder.mkdir(parents=True, exist_ok=True)

            sdfg = dace.SDFG.from_json(inp.program_source.source_code)
            sdfg.build_folder = sdfg_build_folder
            with locking.lock(sdfg_build_folder):
                sdfg_program = sdfg.compile(validate=False)

        assert inp.binding_source is not None
        return CompiledDaceProgram(
            sdfg_program,
            self.bind_func_name,
            inp.binding_source,
        )


class DaCeCompilationStepFactory(factory.Factory):
    class Meta:
        model = DaCeCompiler



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import enum
from typing import ClassVar, Optional, Type

from gt4py import eve
from gt4py.next.iterator import ir
from gt4py.next.iterator.type_system import inference as itir_type_inference


class FixedPointTransformation(eve.NodeTranslator):
    """
    Base class for iterative transformations that converge when a fixed-point is reached.
    """

    REINFER_TYPES: ClassVar[bool] = False

    def visit(self, node, **kwargs):
        node = super().visit(node, **kwargs)
        return self.fp_transform(node, **kwargs) if isinstance(node, ir.Node) else node

    def fp_transform(self, node: ir.Node, **kwargs) -> ir.Node:
        """
        Transform node until a fixed point is reached, e.g. no transformation is applicable anymore.
        """
        while True:
            new_node = self.transform(node, **kwargs)
            if new_node is None:
                break
            else:
                new_node = self._post_transform(node, new_node)
            assert new_node != node
            node = new_node
        return node

    def _post_transform(self, node: ir.Node, new_node: ir.Node) -> ir.Node:
        if self.REINFER_TYPES:
            itir_type_inference.reinfer(new_node)
        self._preserve_annex(node, new_node)
        return new_node

    def transform(self, node: ir.Node, **kwargs) -> Optional[ir.Node]:
        """
        Transform node once.

        Execute transformation if applicable. When a transformation occurred the function will return
        the transformed node. Note that the transformation itself may call other transformations on
        child nodes again.
        """
        ...


@dataclasses.dataclass(frozen=True, kw_only=True)
class CombinedFixedPointTransform(FixedPointTransformation):
    """
    Base class for a set of iterative transformations that converge when a fixed-point is reached.
    """

    #: Enum of all transformation (names). The transformations need to be defined as methods
    #: named `transform_<NAME>`.
    Transformation: ClassVar[Type[enum.Flag]]

    #: All transformations enabled in this instance, e.g. `Transformation.T1 & Transformation.T2`.
    #: Usually the default value is chosen to be all transformations.
    enabled_transformations: enum.Flag

    def transform(self, node: ir.Node, **kwargs) -> Optional[ir.Node]:
        for transformation in self.Transformation:
            if self.enabled_transformations & transformation:
                assert isinstance(transformation.name, str)
                method = getattr(self, f"transform_{transformation.name.lower()}")
                result = method(node, **kwargs)
                if result is not None:
                    assert result is not node, (
                        f"Transformation {transformation.name.lower()} should have returned None, since nothing changed."
                    )
                    return result
        return None



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import re
from typing import Final, Literal

import dace

from gt4py.next import common as gtx_common
from gt4py.next.iterator import builtins as gtir_builtins
from gt4py.next.iterator.ir_utils import ir_makers as im
from gt4py.next.program_processors.runners.dace.lowering import gtir_python_codegen
from gt4py.next.type_system import type_specifications as ts


# arrays for connectivity tables use the following prefix
CONNECTIVITY_INDENTIFIER_PREFIX: Final[str] = "gt_conn_"
CONNECTIVITY_INDENTIFIER_RE: Final[re.Pattern] = re.compile(r"^gt_conn_(\S+)$")

# regex for field size/stride symbol name
FIELD_SYMBOL_RE: Final[re.Pattern] = re.compile(r"^__(\S+)_(\S+)_(size|stride)$")

# element data type for field size/stride symbols
FIELD_SYMBOL_DTYPE: Final[dace.typeclass] = getattr(dace, gtir_builtins.INTEGER_INDEX_BUILTIN)


def as_dace_type(type_: ts.ScalarType) -> dace.typeclass:
    """Converts GT4Py scalar type to corresponding DaCe type."""

    match type_.kind:
        case ts.ScalarKind.BOOL:
            return dace.bool_
        case ts.ScalarKind():
            return getattr(dace, type_.kind.name.lower())
        case _:
            raise ValueError(f"Scalar type '{type_}' not supported.")


def as_itir_type(dtype: dace.typeclass) -> ts.ScalarType:
    """Get GT4Py scalar representation of a DaCe type."""
    type_name = str(dtype.as_numpy_dtype())
    try:
        kind = getattr(ts.ScalarKind, type_name.upper())
    except AttributeError as ex:
        raise ValueError(f"Data type {type_name} not supported.") from ex
    return ts.ScalarType(kind)


def connectivity_identifier(name: str) -> str:
    return f"{CONNECTIVITY_INDENTIFIER_PREFIX}{name}"


def is_connectivity_identifier(
    name: str, offset_provider_type: gtx_common.OffsetProviderType | None = None
) -> bool:
    if (m := CONNECTIVITY_INDENTIFIER_RE.match(name)) is None:
        return False
    elif offset_provider_type is None:
        # If no offset provider type is provided, we assume there is a connectivity identifier
        # that matches the CONNECTIVITY_INDENTIFIER_RE.
        return True
    else:
        return gtx_common.has_offset(offset_provider_type, m[1])


def _field_symbol(
    field_name: str,
    dim: gtx_common.Dimension,
    sym: Literal["size", "stride"],
    offset_provider_type: gtx_common.OffsetProviderType | None,
) -> dace.symbol:
    if (m := CONNECTIVITY_INDENTIFIER_RE.match(field_name)) is None:
        name = f"__{field_name}_{dim.value}_{sym}"
    else:  # a connectivity field
        assert offset_provider_type is not None
        assert m[1] in offset_provider_type
        offset = m[1]
        conn_type = offset_provider_type[offset]
        assert isinstance(conn_type, gtx_common.NeighborConnectivityType)
        if dim == conn_type.source_dim:
            name = f"__{field_name}_source_{sym}"
        elif dim == conn_type.neighbor_dim:
            name = f"__{field_name}_neighbor_{sym}"
        else:
            raise ValueError(f"Unexpect dimension '{dim}' for '{offset}' connectivity.")
    return dace.symbol(name, FIELD_SYMBOL_DTYPE)


def field_size_symbol(
    field_name: str,
    dim: gtx_common.Dimension,
    offset_provider_type: gtx_common.OffsetProviderType,
) -> dace.symbol:
    return _field_symbol(field_name, dim, "size", offset_provider_type)


def field_stride_symbol(
    field_name: str,
    dim: gtx_common.Dimension,
    offset_provider_type: gtx_common.OffsetProviderType | None = None,
) -> dace.symbol:
    return _field_symbol(field_name, dim, "stride", offset_provider_type)


def _range_symbol_name(field_name: str, axis: str) -> str:
    """Common part of the name for the range start/stop symbols."""
    dim = gtx_common.Dimension(axis)
    field_range = im.call("get_domain_range")(field_name, dim)
    return gtir_python_codegen.get_source(field_range)


def range_start_symbol(field_name: str, dim: gtx_common.Dimension) -> dace.symbol:
    """Format name of the start symbol for domain range."""
    name = f"{_range_symbol_name(field_name, dim.value)}_0"
    return dace.symbol(name, FIELD_SYMBOL_DTYPE)


def range_stop_symbol(field_name: str, dim: gtx_common.Dimension) -> dace.symbol:
    """Format name of the stop symbol for domain range."""
    name = f"{_range_symbol_name(field_name, dim.value)}_1"
    return dace.symbol(name, FIELD_SYMBOL_DTYPE)


def filter_connectivity_types(
    offset_provider_type: gtx_common.OffsetProviderType,
) -> dict[str, gtx_common.NeighborConnectivityType]:
    """
    Filter offset provider types of type `NeighborConnectivityType`.

    In other words, filter out the cartesian offset providers.
    """
    return {
        offset: conn
        for offset, conn in offset_provider_type.items()
        if isinstance(conn, gtx_common.NeighborConnectivityType)
    }



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.transforms import symbol_ref_utils


class InlineFundefs(PreserveLocationVisitor, NodeTranslator):
    def visit_SymRef(self, node: itir.SymRef, *, symtable: Dict[str, Any]):
        if node.id in symtable and isinstance(
            (symbol := symtable[node.id]), itir.FunctionDefinition
        ):
            return itir.Lambda(
                params=self.generic_visit(symbol.params, symtable=symtable),
                expr=self.generic_visit(symbol.expr, symtable=symtable),
            )
        return self.generic_visit(node)

    def visit_Program(self, node: itir.Program):
        return self.generic_visit(node, symtable=node.annex.symtable)


def prune_unreferenced_fundefs(program: itir.Program) -> itir.Program:
    """
    Remove all function declarations that are never called.

    >>> from gt4py.next import common
    >>> from gt4py.next.iterator.ir_utils import ir_makers as im
    >>> fun1 = itir.FunctionDefinition(
    ...     id="fun1",
    ...     params=[im.sym("a")],
    ...     expr=im.deref("a"),
    ... )
    >>> fun2 = itir.FunctionDefinition(
    ...     id="fun2",
    ...     params=[im.sym("a")],
    ...     expr=im.deref("a"),
    ... )
    >>> IDim = common.Dimension(value="IDim", kind=common.DimensionKind.HORIZONTAL)
    >>> program = itir.Program(
    ...     id="testee",
    ...     function_definitions=[fun1, fun2],
    ...     params=[im.sym("inp"), im.sym("out")],
    ...     declarations=[],
    ...     body=[
    ...         itir.SetAt(
    ...             expr=im.call("fun1")("inp"),
    ...             domain=im.domain("cartesian_domain", {IDim: (0, 10)}),
    ...             target=im.ref("out"),
    ...         )
    ...     ],
    ... )
    >>> print(prune_unreferenced_fundefs(program))
    testee(inp, out) {
      fun1 = λ(a) → ·a;
      out @ c⟨ IDimₕ: [0, 10[ ⟩ ← fun1(inp);
    }
    """
    fun_names = [fun.id for fun in program.function_definitions]
    referenced_fun_names = symbol_ref_utils.collect_symbol_refs(program.body, fun_names)

    new_fun_defs = []
    for fun_def in program.function_definitions:
        if fun_def.id in referenced_fun_names:
            new_fun_defs.append(fun_def)

    return itir.Program(
        id=program.id,
        function_definitions=new_fun_defs,
        params=program.params,
        declarations=program.declarations,
        body=program.body,
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, TypeAlias

import dace
from dace import data as dace_data
from dace.sdfg import nodes as dace_nodes

from gt4py.next import common as gtx_common
from gt4py.next.program_processors.runners.dace import (
    sdfg_args as gtx_dace_args,
    transformations as gtx_transformations,
)


PropagatedStrideRecord: TypeAlias = tuple[str, dace_nodes.NestedSDFG]
"""Record of a stride that has been propagated into a NestedSDFG.

The type combines the NestedSDFG into which the strides were already propagated
and the data within that NestedSDFG to which we have propagated the strides,
which is the connector name on the NestedSDFG.
We need the NestedSDFG because we have to know what was already processed,
however, we also need the inner array name because of aliasing, i.e. a data
descriptor on the outside could be mapped to multiple data descriptors
inside the NestedSDFG.
"""


def gt_change_strides(
    sdfg: dace.SDFG,
    prefered_direction_kind: gtx_common.DimensionKind,
) -> dace.SDFG:
    """Modifies the strides of transients.

    The function will analyse the access patterns and set the strides of
    transients in the optimal way. The function should run after _all_
    Maps have been created.
    After the adjustment of the strides they will be propagated into the nested
    SDFGs, see `gt_propagate_strides_of()` for more.

    Args:
        sdfg: The SDFG to process.
        prefered_direction_kind: `DimensionKind` of the dimension with stride 1.

    Note:
        - This function should be run after `gt_set_iteration_order()` has been run.
        - Currently the function will not scan the access pattern. Instead it relies
            on the default behaviour of the lowering and how the GT4Py allocator works.
        - The current implementation assumes that there is only one dimension of the
            given kind.
    """
    # TODO(phimeull): Implement this function correctly, such that it decides the
    #   order based on the access pattern. Probably also merge it with
    #   `gt_set_iteration_order()` as the two things are related.

    # NOTE: This function builds inherently assumes the dimension order defined by
    #   `gtx_common.order_dimensions()`, the default behaviour of the lowering,
    #   partially how the  GT4Py allocator works and that there is only one dimension
    #   of any kind (which is true for ICON4Py, but not in general, for example
    #   in Cartesian grids). Its base assumption is that the ordering (Map parameters
    #   and strides) generated by the lowering "out of the box" are in row major/C
    #   order. Because of the GT4Py dimension order this is the right order for
    #   `gtx_common.DimensionKind.VERTICAL`. If the primary direction kind is
    #   `HORIZONTAL`, then according to the GT4Py dimension order column major/FORTRAN
    #   order should be used. To get there we have to reverse the strides order, which
    #   `_gt_change_strides_non_recursive_impl()` does. This is very brittle but at
    #   this point the best thing we can do.

    match prefered_direction_kind:
        case gtx_common.DimensionKind.VERTICAL:
            return  # Nothing to do in that case. Maybe run Memlet propagation here?

        case gtx_common.DimensionKind.HORIZONTAL:
            for nsdfg in sdfg.all_sdfgs_recursive():
                _gt_change_strides_non_recursive_impl(nsdfg)

        case _:
            raise ValueError(
                f"Encountered unknown `DimensionKind` value: {prefered_direction_kind}"
            )


def _gt_change_strides_non_recursive_impl(
    sdfg: dace.SDFG,
) -> None:
    """Set "optimal" strides of all access nodes in the SDFG.

    The function will look for all top level access node, see `_gt_find_toplevel_data_accesses()`
    and set their strides such that the access is optimal, see Note. The function
    will also run `gt_propagate_strides_of()` to propagate the strides into nested SDFGs.

    This function should never be called directly but always through `gt_change_strides()`!

    Note:
        This function has the same underlying assumption as they are outlined in
        `gt_change_strides()`, see there from more informations about the underlying
        assumptions and limitations.
    """
    # NOTE: We have to process all access nodes (transient and globals). If we are inside a
    #   NestedSDFG then they were handled before on the level above us.
    top_level_transients_and_their_accesses = _gt_find_toplevel_data_accesses(
        sdfg=sdfg,
        only_transients=False,
        only_arrays=True,
    )
    for top_level_transient, accesses in top_level_transients_and_their_accesses.items():
        desc: dace_data.Array = sdfg.arrays[top_level_transient]

        # Setting the strides only make sense if we have more than one dimensions
        ndim = len(desc.shape)
        if ndim <= 1:
            continue

        # We assume that everything is in C order initially, to get FORTRAN order
        #  we simply have to reverse the order. This is necessary only for transient
        #  access nodes because the non-transients come from outside and have their
        #  own strides.
        # TODO(phimuell): Set the stride based on the actual access pattern.
        if desc.transient:
            new_stride_order = list(range(ndim))
            desc.set_strides_from_layout(*new_stride_order)

        # Now we have to propagate the changed strides. Because we already have
        #  collected all the AccessNodes we are using the
        #  `gt_propagate_strides_from_access_node()` function, but we have to
        #  create `processed_nsdfg` set already outside here.
        #  While global access nodes in top level SDFGs should have the correct strides,
        #  we need to propagate those strides into the nested SDFGs that use them.
        # TODO(phimuell): Updated the functions such that only one scan is needed.
        processed_nsdfgs: set[dace_nodes.NestedSDFG] = set()
        for state, access_node in accesses:
            gt_propagate_strides_from_access_node(
                sdfg=sdfg,
                state=state,
                outer_node=access_node,
                processed_nsdfgs=processed_nsdfgs,
            )

    # Now handle the views.
    # TODO(phimuell): Remove once `gt_propagate_strides_from_access_node()` can
    #   handle views. However, we should get to a point where we do not have views.
    _gt_modify_strides_of_views_non_recursive(sdfg)


def gt_propagate_strides_of(sdfg: dace.SDFG, data_name: str) -> None:
    """Propagates the strides of `data_name` within the whole SDFG.

    This function will call `gt_propagate_strides_from_access_node()` for every
    AccessNode that refers to `data_name`. It will also make sure that a descriptor
    inside a NestedSDFG is only processed once.

    Args:
        sdfg: The SDFG on which we operate.
        data_name: Name of the data descriptor that should be handled.
    """

    # Defining it here ensures that we will not enter an NestedSDFG multiple times.
    processed_nsdfgs: set[PropagatedStrideRecord] = set()

    for state in sdfg.states():
        for dnode in state.data_nodes():
            if dnode.data != data_name:
                continue
            gt_propagate_strides_from_access_node(
                sdfg=sdfg,
                state=state,
                outer_node=dnode,
                processed_nsdfgs=processed_nsdfgs,
            )


def gt_propagate_strides_from_access_node(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    outer_node: dace_nodes.AccessNode,
    processed_nsdfgs: Optional[set[PropagatedStrideRecord]] = None,
) -> None:
    """Propagates the stride of `outer_node` to any adjacent NestedSDFG.

    The function will propagate the strides of the data descriptor `outer_node`
    refers to along all adjacent edges of `outer_node`. If one of these edges
    leads to a NestedSDFG then the function will modify the strides of data
    descriptor within to match the strides on the outside. The function will then
    recursively process NestedSDFG.

    It is important that this function will only handle the NestedSDFGs that are
    reachable from `outer_node`. To fully propagate the strides the
    `gt_propagate_strides_of()` should be used.

    Args:
        sdfg: The SDFG to process.
        state: The state where the data node is used.
        edge: The edge that reads from the data node, the nested SDFG is expected as the destination.
        outer_node: The data node whose strides should be propagated.
        processed_nsdfgs: Set of NestedSDFG that were already processed and will be ignored.
            Only specify when you know what your are doing.

    Todo:
        Once this function supports the propagation to views `_gt_modify_strides_of_views()`
        can be removed.
    """
    assert isinstance(state, dace.SDFGState)

    if processed_nsdfgs is None:
        # For preventing the case that nested SDFGs are handled multiple time.
        processed_nsdfgs = set()

    for in_edge in state.in_edges(outer_node):
        gt_map_strides_to_src_nested_sdfg(
            sdfg=sdfg,
            state=state,
            edge=in_edge,
            outer_node=outer_node,
            processed_nsdfgs=processed_nsdfgs,
        )
    for out_edge in state.out_edges(outer_node):
        gt_map_strides_to_dst_nested_sdfg(
            sdfg=sdfg,
            state=state,
            edge=out_edge,
            outer_node=outer_node,
            processed_nsdfgs=processed_nsdfgs,
        )


def gt_map_strides_to_dst_nested_sdfg(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge: dace.sdfg.graph.Edge,
    outer_node: dace.nodes.AccessNode,
    processed_nsdfgs: Optional[set[PropagatedStrideRecord]] = None,
) -> None:
    """Propagates the strides of `outer_node` along `edge` in the dataflow direction.

    In this context "along the dataflow direction" means that `edge` is an outgoing
    edge of `outer_node` and the strides are propagated into all NestedSDFGs that
    are downstream of `outer_node`.

    Except in certain cases this function should not be used directly. It is
    instead recommended to use `gt_propagate_strides_of()`, which propagates
    all edges in the SDFG.

    Args:
        sdfg: The SDFG to process.
        state: The state where the data node is used.
        edge: The edge that writes to the data node, the nested SDFG is expected as the source.
        outer_node: The data node whose strides should be propagated.
        processed_nsdfgs: Set of NestedSDFGs that were already processed. Only specify when
            you know what your are doing.
    """
    assert edge.src is outer_node
    _gt_map_strides_to_nested_sdfg_src_dst(
        sdfg=sdfg,
        state=state,
        edge=edge,
        outer_node=outer_node,
        processed_nsdfgs=processed_nsdfgs,
        propagate_along_dataflow=True,
    )


def gt_map_strides_to_src_nested_sdfg(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge: dace.sdfg.graph.Edge,
    outer_node: dace.nodes.AccessNode,
    processed_nsdfgs: Optional[set[PropagatedStrideRecord]] = None,
) -> None:
    """Propagates the strides of `outer_node` along `edge` in the opposite direction of the dataflow

    In this context "in the opposite direction of the dataflow" means that `edge`
    is an incoming edge of `outer_node` and the strides are propagated into all
    NestedSDFGs that are upstream of `outer_node`.

    Except in certain cases this function should not be used directly. It is
    instead recommended to use `gt_propagate_strides_of()`, which propagates
    all edges in the SDFG.

    Args:
        sdfg: The SDFG to process.
        state: The state where the data node is used.
        edge: The edge that writes to the data node, the nested SDFG is expected as the source.
        outer_node: The data node whose strides should be propagated.
        processed_nsdfgs: Set of NestedSDFGs that were already processed. Only specify when
            you know what your are doing.
    """
    _gt_map_strides_to_nested_sdfg_src_dst(
        sdfg=sdfg,
        state=state,
        edge=edge,
        outer_node=outer_node,
        processed_nsdfgs=processed_nsdfgs,
        propagate_along_dataflow=False,
    )


def _gt_map_strides_to_nested_sdfg_src_dst(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
    outer_node: dace.nodes.AccessNode,
    processed_nsdfgs: Optional[set[PropagatedStrideRecord]],
    propagate_along_dataflow: bool,
) -> None:
    """Propagates the stride of `outer_node` along `edge`.

    The function will follow `edge`, the direction depends on the value of
    `propagate_along_dataflow` and propagate the strides of `outer_node`
    into every NestedSDFG that is reachable by following `edge`.

    When the function encounters a NestedSDFG it will determine what data
    the `outer_node` is mapped to on the inside of the NestedSDFG.
    It will then replace the stride of the inner descriptor with the ones
    of the outside. Afterwards it will recursively propagate the strides
    inside the NestedSDFG.
    During this propagation the function will follow any edges.

    If the function reaches a NestedSDFG that is listed inside `processed_nsdfgs`
    then it will be skipped. NestedSDFGs that have been processed will be added
    to the `processed_nsdfgs`.

    Args:
        sdfg: The SDFG to process.
        state: The state where the data node is used.
        edge: The edge that reads from the data node, the nested SDFG is expected as the destination.
        outer_node: The data node whose strides should be propagated.
        processed_nsdfgs: Set of Nested SDFG that were already processed and will be ignored.
            Only specify when you know what your are doing.
        propagate_along_dataflow: Determine the direction of propagation. If `True` the
            function follows the dataflow.

    Note:
        A user should not use this function directly, instead `gt_propagate_strides_of()`,
        `gt_map_strides_to_src_nested_sdfg()` (`propagate_along_dataflow == `False`)
        or `gt_map_strides_to_dst_nested_sdfg()` (`propagate_along_dataflow == `True`)
        should be used.

    Todo:
        Try using `MemletTree` for the propagation.
    """
    # If `processed_nsdfg` is `None` then this is the first call. We will now
    #  allocate the `set` and pass it as argument to all recursive calls, this
    #  ensures that the `set` is the same everywhere.
    if processed_nsdfgs is None:
        processed_nsdfgs = set()

    if propagate_along_dataflow:
        # Propagate along the dataflow or forward, so we are interested at the `dst` of the edge.
        ScopeNode = dace_nodes.MapEntry

        def get_node(edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]) -> dace_nodes.Node:
            return edge.dst

        def get_inner_data(edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]) -> str:
            return edge.dst_conn

        def get_subset(
            state: dace.SDFGState,
            edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
        ) -> dace.subsets.Subset:
            return edge.data.get_src_subset(edge, state)

        def next_edges_by_connector(
            state: dace.SDFGState,
            edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
        ) -> list[dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]]:
            if edge.dst_conn is None or not edge.dst_conn.startswith("IN_"):
                return []
            return list(state.out_edges_by_connector(edge.dst, "OUT_" + edge.dst_conn[3:]))

    else:
        # Propagate against the dataflow or backward, so we are interested at the `src` of the edge.
        ScopeNode = dace_nodes.MapExit

        def get_node(edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]) -> dace_nodes.Node:
            return edge.src

        def get_inner_data(edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]) -> str:
            return edge.src_conn

        def get_subset(
            state: dace.SDFGState,
            edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
        ) -> dace.subsets.Subset:
            return edge.data.get_dst_subset(edge, state)

        def next_edges_by_connector(
            state: dace.SDFGState,
            edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
        ) -> list[dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]]:
            return list(state.in_edges_by_connector(edge.src, "IN_" + edge.src_conn[4:]))

    if isinstance(get_node(edge), ScopeNode):
        for next_edge in next_edges_by_connector(state, edge):
            _gt_map_strides_to_nested_sdfg_src_dst(
                sdfg=sdfg,
                state=state,
                edge=next_edge,
                outer_node=outer_node,
                processed_nsdfgs=processed_nsdfgs,
                propagate_along_dataflow=propagate_along_dataflow,
            )

    elif isinstance(get_node(edge), dace.nodes.NestedSDFG):
        nsdfg_node = get_node(edge)
        inner_data = get_inner_data(edge)
        process_record = (inner_data, nsdfg_node)

        if process_record in processed_nsdfgs:
            # We already handled this NestedSDFG and the inner data.
            return

        # Mark this nested SDFG as processed.
        processed_nsdfgs.add(process_record)

        # Now set the stride of the data descriptor inside the nested SDFG to
        #  the ones it has outside.
        _gt_map_strides_into_nested_sdfg(
            sdfg=sdfg,
            nsdfg_node=nsdfg_node,
            inner_data=inner_data,
            outer_subset=get_subset(state, edge),
            outer_desc=outer_node.desc(sdfg),
        )

        # Since the function call above is not recursive we have now to propagate
        #  the change into the NestedSDFGs. Using `_gt_find_toplevel_data_accesses()`
        #  is a bit overkill, but allows for a more uniform processing.
        # TODO(phimuell): Instead of scanning every level for every data we modify
        #   we should scan the whole SDFG once and then reuse this information.
        accesses_in_nested_sdfg = _gt_find_toplevel_data_accesses(
            sdfg=nsdfg_node.sdfg,
            only_transients=False,  # Because on the nested levels they are globals.
            only_arrays=True,
        )
        for nested_state, nested_access in accesses_in_nested_sdfg.get(inner_data, list()):
            # We have to use `gt_propagate_strides_from_access_node()` here because we
            #  have to handle its entirety. We could wait until the other branch processes
            #  the nested SDFG, but this might not work, so let's do it fully now.
            gt_propagate_strides_from_access_node(
                sdfg=nsdfg_node.sdfg,
                state=nested_state,
                outer_node=nested_access,
                processed_nsdfgs=processed_nsdfgs,
            )


def _gt_map_strides_into_nested_sdfg(
    sdfg: dace.SDFG,
    nsdfg_node: dace.nodes.NestedSDFG,
    inner_data: str,
    outer_subset: dace.subsets.Subset,
    outer_desc: dace_data.Data,
) -> None:
    """Modify the strides of `inner_data` inside `nsdfg_node` to match `outer_desc`.

    `inner_data` is the name of a data descriptor inside the NestedSDFG.
    The function will then modify the strides of `inner_data`, assuming this
    is an array, to match the ones of `outer_desc`.

    Args:
        sdfg: The SDFG containing the NestedSDFG.
        nsdfg_node: The node in the parent SDFG that contains the NestedSDFG.
        inner_data: The name of the data descriptor that should be processed
            inside the NestedSDFG (by construction also a connector name).
        outer_subset: The subset that describes what part of the outer data is
            mapped into the NestedSDFG.
        outer_desc: The data descriptor of the data on the outside.

    Todo:
        - Handle explicit dimensions of size 1.
        - What should we do if the stride symbol is used somewhere else, creating an
            alias is probably not the right thing?
        - Handle the case if the outer stride symbol is already used in another
            context inside the Neste SDFG.
    """
    # We need to compute the new strides. In the following we assume that the
    #  relative order of the dimensions does not change, but we support the case
    #  where some dimensions of the outer data descriptor are not present on the
    #  inside. For example this happens for the Memlet `a[__i0, 0:__a_size1]`. We
    #  detect this case by checking if the Memlet subset in that dimension has size 1.
    # TODO(phimuell): Handle the case were some additional size 1 dimensions are added.
    inner_desc: dace_data.Data = nsdfg_node.sdfg.arrays[inner_data]
    inner_shape = inner_desc.shape

    outer_shape = outer_desc.shape
    outer_strides = outer_desc.strides
    outer_inflow = outer_subset.size()

    if isinstance(inner_desc, dace_data.Scalar):
        # A scalar does not have a stride that must be propagated.
        return

    # Now determine the new stride that is needed on the inside.
    new_strides: list = []
    if len(outer_shape) == len(inner_shape):
        # The inner and the outer descriptor have the same dimensionality.
        #  We now have to decide if we should take the stride from the outside,
        #  which happens for example in case of `A[0:N, 0:M] -> B[N, M]`, or if we
        #  must take 1, which happens if we do `A[0:N, i] -> B[N, 1]`, we detect that
        #  based on the volume that flows in.
        for dim_ostride, dim_oinflow in zip(outer_strides, outer_inflow, strict=True):
            new_strides.append(1 if dim_oinflow == 1 else dim_ostride)

    elif len(inner_shape) < len(outer_shape):
        # There are less dimensions on the inside than on the outside. This means
        #  that some were sliced away. We detect this case by checking if the Memlet
        #  subset in that dimension has size 1.
        #  NOTE: That this is not always correct as it might be possible that there
        #   are some explicit size 1 dimensions at several places.
        new_strides = []
        for dim_ostride, dim_oinflow in zip(outer_strides, outer_inflow, strict=True):
            if dim_oinflow == 1:
                pass
            else:
                new_strides.append(dim_ostride)
            assert len(new_strides) <= len(inner_shape)
    else:
        # The case that we have more dimensions on the inside than on the outside.
        #  This is currently not supported.
        raise NotImplementedError("NestedSDFGs can not be used to increase the rank.")

    if len(new_strides) != len(inner_shape):
        raise ValueError("Failed to compute the inner strides.")

    # For the strides of the arrays inside the nested SDFG we will create a new unique
    #  symbol which is initialized, through the symbol mapping, to the value of this
    #  stride on the outside. The benefit is that only the mapped container is affected
    #  and nothing else. Consider for example the case where initially two arrays
    #  inside the nested SDFG use the same stride symbol, but only one array is mapped.
    #  The main drawback is that the logical connection is lost, thus if the old
    #  stride symbol is used somewhere inside the nested SDFG, with the expectation
    #  that it corresponds to the stride of the inner container, then this connection
    #  is lost. However, this is probably not much of an issue for the strides, but
    #  more problematic for the shape, whose symbols are likely to appear as loop bounds.
    for i, dim_ostride in enumerate(new_strides):
        if str(dim_ostride).isdigit():
            # A literal stride (e.g. `1`) can be set directly
            new_strides[i] = dim_ostride
        else:
            if dim_ostride.is_symbol:
                # Try reusing the same symbol name as the outer stride, but find a new name if already used.
                dim_istride = nsdfg_node.sdfg.add_symbol(
                    dim_ostride.name, sdfg.symbols[dim_ostride.name], find_new_name=True
                )
            else:
                # Map a symbolic expression such as `value1 - value2` to a new stride symbol.
                dim_istride = nsdfg_node.sdfg.add_symbol(
                    f"__{inner_data}_stride_{i}",
                    gtx_dace_args.FIELD_SYMBOL_DTYPE,
                    find_new_name=True,
                )
            new_strides[i] = dace.symbolic.pystr_to_symbolic(dim_istride)
            nsdfg_node.symbol_mapping[dim_istride] = dim_ostride

    # We have to replace the `strides` attribute of the inner descriptor.
    inner_desc.set_shape(inner_desc.shape, new_strides)


def _gt_find_toplevel_data_accesses(
    sdfg: dace.SDFG,
    only_transients: bool,
    only_arrays: bool = False,
) -> dict[str, list[tuple[dace.SDFGState, dace_nodes.AccessNode]]]:
    """Find all data that is accessed on the top level.

    The function will scan the SDFG, ignoring nested one, and return the
    name of all data that only have AccessNodes on the top level. In data
    is found that has an AccessNode on both the top level and in a nested
    scope and error is generated.
    By default the function will return transient and non transient data,
    however, if `only_transients` is `True` then only transient data will
    be returned. In addition by seeting `only_arrays` only arrays will be
    returned.

    Args:
        sdfg: The SDFG to process.
        only_transients: If `True` only include transients.
        only_arrays: If `True`, defaults to `False`, only arrays are returned.
        inlclude_views: Also returns Views.

    Returns:
        A `dict` that maps the name of a data container, to a list of tuples
        containing the state where the AccessNode was found and the AccessNode.
    """
    # List of data that is accessed on the top level and all its access node.
    top_level_data: dict[str, list[tuple[dace.SDFGState, dace_nodes.AccessNode]]] = dict()

    # List of all data that were found not on top level.
    not_top_level_data: set[str] = set()

    for state in sdfg.states():
        assert isinstance(state, dace.SDFGState)
        scope_dict = state.scope_dict()
        for dnode in state.data_nodes():
            data: str = dnode.data
            if scope_dict[dnode] is not None:
                # The node was not found on the top level. So we can ignore it.
                # We also check if it was ever found on the top level, this should
                #  not happen, as everything should go through Maps. But some strange
                #  DaCe transformation might do it.
                assert data not in top_level_data, (
                    f"Found {data} on the top level and inside a scope."
                )
                not_top_level_data.add(data)
                continue

            elif data in top_level_data:
                # The data is already known to be in top level data, so we must add the
                #  AccessNode to the list of known nodes. But nothing else.
                top_level_data[data].append((state, dnode))
                continue

            elif gtx_transformations.utils.is_view(dnode, sdfg):
                # The AccessNode refers to a View so we ignore it anyway.
                continue

            # We have found a new data node that is on the top node and is unknown.
            assert data not in not_top_level_data, (
                f"Found {data} on the top level and inside a scope."
            )
            desc: dace_data.Data = dnode.desc(sdfg)

            # Check if we only accept arrays
            if only_arrays and not isinstance(desc, dace_data.Array):
                continue

            # For now we ignore registers.
            #  We do this because register are allocated on the stack, so the compiler
            #  has all information and should organize the best thing possible.
            # TODO(phimuell): verify this.
            elif desc.storage is dace.StorageType.Register:
                continue

            # We are only interested in transients
            if only_transients and (not desc.transient):
                continue

            # Now create the new entry in the list and record the AccessNode.
            top_level_data[data] = [(state, dnode)]
    return top_level_data


def _gt_modify_strides_of_views_non_recursive(sdfg: dace.SDFG) -> None:
    """The function determines the strides of Views.

    The function should not be called directly, instead it is called by
    `gt_change_strides()` directly if needed. The function will recursively
    process the SDFG and modifies the strides.

    Todo:
        Once `gt_propagate_strides_from_access_node()` can handle View updates
        evaluate if this function is still needed.
    """
    for state in sdfg.states():
        scope_dict = state.scope_dict()
        propagation_record: set[PropagatedStrideRecord] = set()
        for view_node in state.data_nodes():
            view_desc = view_node.desc(sdfg)
            if not isinstance(view_desc, dace_data.View):
                continue
            viewed_node = gtx_transformations.utils.track_view(view_node, state, sdfg)
            viewed_desc = viewed_node.desc(sdfg)

            # We are not able to handle tower of Views
            if isinstance(viewed_desc, dace_data.View):
                raise NotImplementedError(
                    f"Can not handle the view '{view_node.data}' that views view '{viewed_node.data}'"
                )

            # If both the View and the viewed node are not on the top level then do
            #  nothing. Why? The answer is that this function is only called by
            #  `_gt_change_strides_non_recursive_impl()` which only manipulates
            #  the strides of transients at the top level and leaves the one inside
            #  a Map alone. Thus there was no modification and no change.
            if scope_dict[viewed_node] is not None and scope_dict[view_node] is not None:
                continue

            # If the viewed data is global data, then we do not modify the strides because
            #  we assume that it was set correctly from the beginning and the viewed strides
            #  have not changed.
            # TODO(phimuell): Check interaction with the demote feature of auto optimizer.
            if not viewed_desc.transient:
                continue

            # There is a special case if the View is a scalar or has a shape of `(1,)`
            #  then strides are meaningless.
            if isinstance(view_desc, dace_data.Scalar) or (
                len(view_desc.shape) == 1 and ((view_desc.shape[0] == 1) == True)  # noqa: E712 [true-false-comparison]  # SymPy comparison
            ):
                continue

            # If the dimensionality of the two data is different then we can not handle it.
            #  This is probably the most difficult case to handle. However, instead of
            #  handling the case, we should simply remove the View.
            if len(view_desc.shape) != len(viewed_desc.shape):
                raise NotImplementedError(
                    f"Can not handle the change from {len(viewed_desc.shape)} ({viewed_node.data}) to {len(view_desc.shape)} ({view_node.data})."
                )

            # Even if they have the same dimensionality, we can not simply copy the strides.
            #  Consider the case were `viewed_desc` is a 2D array with the following shape
            #  `(N, M)`. `view_desc` is a vertical slice, but for some reasons has an
            #  additional dummy dimension, i.e. its shape is either `(N, 1)` or `(1, N)`.
            #  So copying the strides around is not gonna work.
            if view_desc.shape != viewed_desc.shape:
                raise NotImplementedError(
                    f"Can not change from shape `{viewed_desc.shape}` ({viewed_node.data})to shape `{view_desc.shape}` ({view_node.data})."
                )

            # In case they have the same shape we are fine.
            view_desc.strides = viewed_desc.strides
            gt_propagate_strides_from_access_node(
                sdfg=sdfg,
                state=state,
                outer_node=view_node,
                processed_nsdfgs=propagation_record,
            )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import functools

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next import utils
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.ir_utils import (
    common_pattern_matcher as cpm,
    domain_utils,
    ir_makers as im,
)
from gt4py.next.iterator.transforms import symbol_ref_utils
from gt4py.next.iterator.type_system import inference as type_inference
from gt4py.next.type_system import type_specifications as ts


def _in(pos: itir.Expr, domain: itir.Expr) -> itir.Expr:
    """
    Given a position and a domain return an expression that evaluates to `True` if the position is inside the domain.

    pos = `{i, j, k}`, domain = `u⟨ Iₕ: [i0, i1[, Iₕ: [j0, j1[, Iₕ: [k0, k1[ ⟩`
    -> `((i0 <= i) & (i < i1)) & ((j0 <= j) & (j < j1)) & ((k0 <= k)l & (k < k1))`
    """
    ret = [
        im.and_(
            im.less_equal(v.start, im.tuple_get(i, pos)),
            im.less(im.tuple_get(i, pos), v.stop),
        )
        for i, v in enumerate(domain_utils.SymbolicDomain.from_expr(domain).ranges.values())
    ]
    return functools.reduce(im.and_, ret)


class _TransformToAsFieldop(PreserveLocationVisitor, NodeTranslator):
    PRESERVED_ANNEX_ATTRS = (
        "type",
        "domain",
    )

    @classmethod
    def apply(cls, node: itir.Node):
        """
        Transform `concat_where` expressions into equivalent `as_fieldop` expressions.

        Note that (backward) domain inference may not be executed after this pass as it can not
        correctly infer the accessed domains when the value selection is represented as an `if_`
        inside the `as_fieldop.
        """
        node = cls().visit(node)
        node = type_inference.SanitizeTypes().visit(node)
        return node

    def visit_FunCall(self, node: itir.FunCall) -> itir.FunCall:
        node = self.generic_visit(node)
        if cpm.is_call_to(node, "concat_where"):
            cond, true_branch, false_branch = node.args
            assert isinstance(cond.type, ts.DomainType)
            position = [im.index(dim) for dim in cond.type.dims]
            refs = symbol_ref_utils.collect_symbol_refs(cond)

            domains: tuple[domain_utils.SymbolicDomain, ...] = utils.flatten_nested_tuple(
                node.annex.domain
            )
            assert all(domain == domains[0] for domain in domains), (
                "At this point all `concat_where` arguments should be posed on the same domain."
            )
            assert isinstance(domains[0], domain_utils.SymbolicDomain)
            domain_expr = domains[0].as_expr()

            return im.as_fieldop(
                im.lambda_("__tcw_pos", "__tcw_arg0", "__tcw_arg1", *refs)(
                    im.let(*zip(refs, map(im.deref, refs), strict=True))(
                        im.if_(
                            _in(im.deref("__tcw_pos"), cond),
                            im.deref("__tcw_arg0"),
                            im.deref("__tcw_arg1"),
                        )
                    )
                ),
                domain_expr,
            )(im.make_tuple(*position), true_branch, false_branch, *refs)

        return node


transform_to_as_fieldop = _TransformToAsFieldop.apply



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause
from __future__ import annotations

import functools
import pickle
import typing
from typing import TYPE_CHECKING, ClassVar, List, Optional, Union

import gt4py.eve as eve
from gt4py.eve import Coerced, SymbolName, SymbolRef, concepts, utils as eve_utils
from gt4py.eve.traits import SymbolTableTrait, ValidatedSymbolTableTrait
from gt4py.next import common
from gt4py.next.iterator.builtins import BUILTINS
from gt4py.next.type_system import type_specifications as ts


DimensionKind = common.DimensionKind


#: A custom pickler which ignores "location" and "type" attribute of IR nodes.
_semantic_node_pickler: type[pickle.Pickler] = concepts.skipping_fields_node_pickler(
    "location", "type"
)

#: Generates an unique fingerprint for IR nodes ignoring their location and type attributes.
semantic_fingerprint: concepts.NodeFingerprinter = functools.partial(
    eve_utils.content_hash, pickler=_semantic_node_pickler
)


@eve_utils.noninstantiable
class Node(eve.Node):
    location: Optional[concepts.SourceLocation] = eve.field(default=None, repr=False, compare=False)

    # TODO(tehrengruber): include in comparison if value is not None
    type: Optional[ts.TypeSpec] = eve.field(default=None, repr=False, compare=False)

    def fingerprint(self) -> str:
        """
        Generates a unique hash string for this node that is location and type agnostic.
        """
        return semantic_fingerprint(self)

    def __str__(self) -> str:
        from gt4py.next.iterator.pretty_printer import pformat

        return pformat(self)

    def __hash__(self) -> int:
        return hash(
            (
                type(self),
                *(
                    tuple(v) if isinstance(v, list) else v
                    for (k, v) in self.iter_children_items()
                    if k not in ["location", "type"]
                ),
            )
        )


class Sym(Node):  # helper
    id: Coerced[SymbolName]


@eve_utils.noninstantiable
class Expr(Node): ...


class Literal(Expr):
    value: str
    type: ts.ScalarType


class NoneLiteral(Expr):
    _none_literal: int = 0


class InfinityLiteral(Expr):
    # TODO(tehrengruber): self referential `ClassVar` not supported in eve.
    if TYPE_CHECKING:
        POSITIVE: ClassVar[InfinityLiteral]
        NEGATIVE: ClassVar[InfinityLiteral]

    name: typing.Literal["POSITIVE", "NEGATIVE"]

    def __str__(self):
        return f"{type(self).__name__}.{self.name}"


InfinityLiteral.NEGATIVE = InfinityLiteral(name="NEGATIVE")
InfinityLiteral.POSITIVE = InfinityLiteral(name="POSITIVE")


class OffsetLiteral(Expr):
    value: Union[int, str]


class AxisLiteral(Expr):
    # TODO(havogt): Refactor to use declare Axis/Dimension at the Program level.
    # Now every use of the literal has to provide the kind, where usually we only care of the name.
    value: str
    kind: common.DimensionKind = common.DimensionKind.HORIZONTAL


class SymRef(Expr):
    id: Coerced[SymbolRef]


class Lambda(Expr, SymbolTableTrait):
    params: List[Sym]
    expr: Expr


class FunCall(Expr):
    fun: Expr  # VType[Callable]
    args: List[Expr]


class FunctionDefinition(Node, SymbolTableTrait):
    id: Coerced[SymbolName]
    params: List[Sym]
    expr: Expr


class Stmt(Node): ...


class SetAt(Stmt):  # from JAX array.at[...].set()
    expr: Expr  # only `as_fieldop(stencil)(inp0, ...)` in first refactoring
    domain: Expr
    target: Expr  # `make_tuple` or SymRef


class IfStmt(Stmt):
    cond: Expr
    true_branch: list[Stmt]
    false_branch: list[Stmt]


class Temporary(Node):
    id: Coerced[eve.SymbolName]
    domain: Optional[Expr] = None
    dtype: Optional[ts.ScalarType | ts.TupleType] = None


class Program(Node, ValidatedSymbolTableTrait):
    id: Coerced[SymbolName]
    function_definitions: List[FunctionDefinition]
    params: List[Sym]
    declarations: List[Temporary]
    body: List[Stmt]

    _NODE_SYMBOLS_: ClassVar[List[Sym]] = [
        Sym(id=name) for name in sorted(BUILTINS)
    ]  # sorted for serialization stability


# TODO(fthaler): just use hashable types in nodes (tuples instead of lists)
Sym.__hash__ = Node.__hash__  # type: ignore[method-assign]
Expr.__hash__ = Node.__hash__  # type: ignore[method-assign]
Literal.__hash__ = Node.__hash__  # type: ignore[method-assign]
NoneLiteral.__hash__ = Node.__hash__  # type: ignore[method-assign]
OffsetLiteral.__hash__ = Node.__hash__  # type: ignore[method-assign]
AxisLiteral.__hash__ = Node.__hash__  # type: ignore[method-assign]
SymRef.__hash__ = Node.__hash__  # type: ignore[method-assign]
Lambda.__hash__ = Node.__hash__  # type: ignore[method-assign]
FunCall.__hash__ = Node.__hash__  # type: ignore[method-assign]
FunctionDefinition.__hash__ = Node.__hash__  # type: ignore[method-assign]
Program.__hash__ = Node.__hash__  # type: ignore[method-assign]
SetAt.__hash__ = Node.__hash__  # type: ignore[method-assign]
IfStmt.__hash__ = Node.__hash__  # type: ignore[method-assign]
InfinityLiteral.__hash__ = Node.__hash__  # type: ignore[method-assign]
Temporary.__hash__ = Node.__hash__  # type: ignore[method-assign]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import functools
from typing import Final

import factory

from gt4py._core import definitions as core_defs, filecache
from gt4py.next import config
from gt4py.next.otf import recipes, stages, workflow
from gt4py.next.program_processors.runners.dace.workflow import (
    bindings as bindings_step,
    decoration as decoration_step,
)
from gt4py.next.program_processors.runners.dace.workflow.compilation import (
    DaCeCompilationStepFactory,
)
from gt4py.next.program_processors.runners.dace.workflow.translation import (
    DaCeTranslationStepFactory,
)


_GT_DACE_BINDING_FUNCTION_NAME: Final[str] = "update_sdfg_args"


class DaCeWorkflowFactory(factory.Factory):
    class Meta:
        model = recipes.OTFCompileWorkflow

    class Params:
        auto_optimize: bool = False
        device_type: core_defs.DeviceType = core_defs.DeviceType.CPU
        cmake_build_type: config.CMakeBuildType = factory.LazyFunction(  # type: ignore[assignment] # factory-boy typing not precise enough
            lambda: config.CMAKE_BUILD_TYPE
        )

        cached_translation = factory.Trait(
            translation=factory.LazyAttribute(
                lambda o: workflow.CachedStep(
                    o.bare_translation,
                    hash_function=stages.fingerprint_compilable_program,
                    cache=filecache.FileCache(str(config.BUILD_CACHE_DIR / "translation_cache")),
                )
            ),
        )

        bare_translation = factory.SubFactory(
            DaCeTranslationStepFactory,
            device_type=factory.SelfAttribute("..device_type"),
            auto_optimize=factory.SelfAttribute("..auto_optimize"),
        )

    translation = factory.LazyAttribute(lambda o: o.bare_translation)
    bindings = factory.LazyAttribute(
        lambda o: functools.partial(
            bindings_step.bind_sdfg,
            bind_func_name=_GT_DACE_BINDING_FUNCTION_NAME,
        )
    )
    compilation = factory.SubFactory(
        DaCeCompilationStepFactory,
        bind_func_name=_GT_DACE_BINDING_FUNCTION_NAME,
        cache_lifetime=factory.LazyFunction(lambda: config.BUILD_CACHE_LIFETIME),
        device_type=factory.SelfAttribute("..device_type"),
        cmake_build_type=factory.SelfAttribute("..cmake_build_type"),
    )
    decoration = factory.LazyAttribute(
        lambda o: functools.partial(
            decoration_step.convert_args,
            device=o.device_type,
        )
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Shared build system functionality."""

from __future__ import annotations

import importlib


def python_module_suffix() -> str:
    return importlib.machinery.EXTENSION_SUFFIXES[0][1:]
